@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
-# Copyright 2017 Yelp
-# Copyright 2018 Yelp
+# Copyright 2017-2018 Yelp
+# Copyright 2019 Yelp
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -20,6 +20,7 @@ import socket
 import random
 import signal
 import time
+from copy import deepcopy
 from os.path import basename
 from subprocess import Popen
 from subprocess import PIPE
@@ -50,6 +51,9 @@ _EXT_TO_UNARCHIVE_CMD = {
 # issue a warning if max_mins_idle is set to less than this
 _DEFAULT_MAX_MINS_IDLE = 10.0
 
+# default part size (so we can share with Spark runner)
+_DEFAULT_CLOUD_PART_SIZE_MB = 100
+
 
 class HadoopInTheCloudJobRunner(MRJobBinRunner):
     """Abstract base class for all Hadoop-in-the-cloud services."""
@@ -72,7 +76,6 @@ class HadoopInTheCloudJobRunner(MRJobBinRunner):
         'instance_type',
         'master_instance_type',
         'max_mins_idle',
-        'max_hours_idle',
         'num_core_instances',
         'num_task_instances',
         'region',
@@ -125,11 +128,12 @@ class HadoopInTheCloudJobRunner(MRJobBinRunner):
 
     ### Options ###
 
-    def _default_opts(self):
+    @classmethod
+    def _default_opts(cls):
         return combine_dicts(
-            super(HadoopInTheCloudJobRunner, self)._default_opts(),
+            super(HadoopInTheCloudJobRunner, cls)._default_opts(),
             dict(
-                cloud_part_size_mb=100,  # 100 MB
+                cloud_part_size_mb=_DEFAULT_CLOUD_PART_SIZE_MB,
                 max_mins_idle=_DEFAULT_MAX_MINS_IDLE,
                 # don't use a list because it makes it hard to read option
                 # values when running in verbose mode. See #1284
@@ -151,16 +155,6 @@ class HadoopInTheCloudJobRunner(MRJobBinRunner):
                               (integer_types, float)):
                 raise TypeError('cloud_part_size_mb must be a number')
 
-        # patch max_hours_idle into max_mins_idle (see #1663)
-        if opts.get('max_hours_idle') is not None:
-            log.warning(
-                'max_hours_idle is deprecated and will be removed in v0.7.0.' +
-                (' Please use max_mins_idle instead'
-                 if opts.get('max_mins_idle') is None else ''))
-
-            if opts.get('max_mins_idle') is None:
-                opts['max_mins_idle'] = opts['max_hours_idle'] * 60
-
         return opts
 
     def _combine_opts(self, opt_list):
@@ -410,9 +404,10 @@ class HadoopInTheCloudJobRunner(MRJobBinRunner):
     def _add_extra_cluster_params(self, params):
         """Return a dict with the *extra_cluster_params* opt patched into
         *params*, and ``None`` values removed."""
-        params = params.copy()
-        params.update(self._opts['extra_cluster_params'])
-        params = {k: v for k, v in params.items() if v is not None}
+        params = deepcopy(params)
+
+        for k, v in sorted(self._opts['extra_cluster_params'].items()):
+            _patch_params(params, k, v)
 
         return params
 
@@ -597,3 +592,29 @@ class HadoopInTheCloudJobRunner(MRJobBinRunner):
             return random.sample(self._opts['ssh_bind_ports'], num_picks)
         finally:
             random.setstate(random_state)
+
+
+def _patch_params(params, name, value):
+    """Helper method for _add_extra_cluster_params().
+
+    Set *name* in *params* to *value*
+
+    If *name* has one or more dots in it, recursively set the value
+    in successive nested dictionaries, creating them if necessary.
+    For example, if *name* is ``Instances.EmrManagedMasterSecurityGroup``,
+    set ``params['Instances']['EmrManagedMasterSecurityGroup']``
+
+    If *value* is ``None``, delete the value (if it exists), rather than
+    setting it to ``None``.
+    """
+    if not isinstance(params, dict):
+        raise TypeError('must be a dictionary')
+
+    if '.' in name:
+        head, rest = name.split('.', 1)
+        _patch_params(params.setdefault(head, {}), rest, value)
+    elif value is None:
+        if name in params:
+            del params[name]
+    else:
+        params[name] = value
