@@ -1,5 +1,7 @@
 # Copyright 2009-2010 Yelp
 # Copyright 2013 David Marin
+# Copyright 2017 Yelp
+# Copyright 2019 Yelp
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,7 +14,6 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-
 """A text classifier that uses a modified version of Naive Bayes that is not
 sensitive to document length.
 
@@ -21,9 +22,13 @@ generally you'd run one job to generate n-gram scores, put them in a sqlite
 database, and run a second job to score documents. But this is simple, and
 it works!
 
-This takes as its input documents encoded by encode_document() below. For each
-document, you specify its text, and whether it belongs or does not belong
-to one or more categories. You can also specify a unique ID for each document.
+This takes text files as its input. Each file's name should have the format
+```unique_id-cat_id_1-cat_id_2-etc.txt```, where ```unique_id``` is unique
+for this document, and ```cat_id_1``` etc. are categories the document
+is known to belong to. You may also use the form ```not_cat_id_3``` to
+indicate that a document is known *not* to belong to ```cat_id_3```. Document
+and category IDs may not contain the ```-``` or ```.``` characters. For
+examples, look in ```docs-to-classify/```.
 
 This job outputs the documents, with the field 'cat_to_score' filled in.
 Generally, positive scores indicate the document is in the category, and
@@ -48,31 +53,36 @@ appears in at least once.
 from collections import defaultdict
 import hashlib
 import math
+import posixpath
 import re
 
 from mrjob.job import MRJob
 from mrjob.protocol import JSONValueProtocol
+from mrjob.py2 import to_unicode
 from mrjob.step import MRStep
+from mrjob.util import file_ext
 
 
-def encode_document(text, cats=None, id=None):
-    """Encode a document as a JSON so that MRTextClassifier can read it.
-
-    Args:
-    text -- the text of the document (as a unicode)
-    cats -- a dictionary mapping a category name (e.g. 'sports') to True if
-        the document is in the category, and False if it's not. None indicates
-        that we have no information about this documents' categories
-    id -- a unique ID for the document (any kind of JSON-able value should
-        work). If not specified, we'll auto-generate one.
+def parse_doc_filename(input_uri):
+    """Parse a filename like ``some_id-cat1-cat2-not_cat3.txt`` into
+    ``dict(id='some_id', cats=dict(cat1=True, cat2=True, cat3=False))``
     """
-    text = unicode(text)
-    cats = dict((unicode(cat), bool(is_in_cat))
-                for cat, is_in_cat
-                in (cats or {}).iteritems())
+    # get filename without extension
+    name_with_ext = posixpath.basename(input_uri)
+    name = name_with_ext[:-len(file_ext(name_with_ext))]
 
-    return JSONValueProtocol.write(
-        None, {'text': text, 'cats': cats, 'id': id}) + '\n'
+    parts = name.split('-')
+
+    doc_id = parts[0]
+    cats = {}
+
+    for part in parts[1:]:
+        if part.startswith('not_'):
+            cats[part[4:]] = False
+        else:
+            cats[part] = True
+
+    return dict(id=doc_id, cats=cats)
 
 
 def count_ngrams(text, max_ngram_size, stop_words):
@@ -129,7 +139,6 @@ DEFAULT_STOP_WORDS = [
 
 
 class MRTextClassifier(MRJob):
-    INPUT_PROTOCOL = JSONValueProtocol
 
     def steps(self):
         """Conceptually, the steps are:
@@ -146,62 +155,63 @@ class MRTextClassifier(MRJob):
         The documents themselves are passed through from step 1 to step 5.
         Ngram scoring information is passed through from step 4 to step 5.
         """
-        return [MRStep(mapper=self.parse_doc,
+        return [MRStep(mapper_raw=self.parse_doc,
                        reducer=self.count_ngram_freq),
                 MRStep(reducer=self.score_ngrams),
                 MRStep(reducer=self.score_documents_by_ngram),
                 MRStep(reducer=self.score_documents)]
 
-    def configure_options(self):
+    def configure_args(self):
         """Add command-line options specific to this script."""
-        super(MRTextClassifier, self).configure_options()
+        super(MRTextClassifier, self).configure_args()
 
-        self.add_passthrough_option(
-            '--min-df', dest='min_df', default=2, type='int',
+        self.add_passthru_arg(
+            '--min-df', dest='min_df', default=2, type=int,
             help=('min number of documents an n-gram must appear in for us to'
-                  ' count it. Default: %default'))
-        self.add_passthrough_option(
-            '--max-df', dest='max_df', default=10000000, type='int',
+                  ' count it. Default: %(default)s'))
+        self.add_passthru_arg(
+            '--max-df', dest='max_df', default=10000000, type=int,
             help=('max number of documents an n-gram may appear in for us to'
                   ' count it (this keeps reducers from running out of memory).'
-                  ' Default: %default'))
-        self.add_passthrough_option(
+                  ' Default: %(default)s'))
+        self.add_passthru_arg(
             '--max-ngram-size', dest='max_ngram_size',
-            default=DEFAULT_MAX_NGRAM_SIZE, type='int',
+            default=DEFAULT_MAX_NGRAM_SIZE, type=int,
             help='maximum phrase length to consider')
-        self.add_passthrough_option(
+        self.add_passthru_arg(
             '--stop-words', dest='stop_words',
             default=', '.join(DEFAULT_STOP_WORDS),
             help=("comma-separated list of words to ignore. For example, "
                   "--stop-words 'in, the' would cause 'hole in the wall' to be"
-                  " parsed as ['hole', 'wall']. Default: %default"))
-        self.add_passthrough_option(
+                  " parsed as ['hole', 'wall']. Default: %(default)s"))
+        self.add_passthru_arg(
             '--short-doc-threshold', dest='short_doc_threshold',
-            type='int', default=None,
+            type=int, default=None,
             help=('Normally, for each n-gram size, we take the average score'
                   ' over all n-grams that appear. This allows us to penalize'
                   ' short documents by using this threshold as the denominator'
                   ' rather than the actual number of n-grams.'))
-        self.add_passthrough_option(
+        self.add_passthru_arg(
             '--no-test-set', dest='no_test_set',
             action='store_true', default=False,
             help=("Choose about half of the documents to be the testing set"
                   " (don't use them to train the classifier) based on a SHA1"
                   " hash of their text"))
 
-    def load_options(self, args):
+    def load_args(self, args):
         """Parse stop_words option."""
-        super(MRTextClassifier, self).load_options(args)
+        super(MRTextClassifier, self).load_args(args)
 
         self.stop_words = set()
         if self.options.stop_words:
             self.stop_words.update(
                 s.strip().lower() for s in self.options.stop_words.split(','))
 
-    def parse_doc(self, _, doc):
+    def parse_doc(self, input_path, input_uri):
         """Mapper: parse documents and emit ngram information.
 
-        Input: JSON-encoded documents (see :py:func:`encode_document`)
+        Input: Text files whose name contains a unique document ID and
+        category information (see :py:func:`parse_doc_filename`).
 
         Output:
         ``('ngram', (n, ngram)), (count, cats)`` OR
@@ -223,34 +233,31 @@ class MRTextClassifier(MRJob):
             in_test_set: boolean indicating if this doc is in the test set
             id: SHA1 hash of doc text (if not already filled)
         """
-        # only compute doc hash if we need it
-        if doc.get('id') is not None and self.options.no_test_set:
-            doc_hash = '0'  # don't need doc hash
-        else:
-            doc_hash = hashlib.sha1(doc['text'].encode('utf-8')).hexdigest()
+        # fill *id* and *cats*
+        doc = parse_doc_filename(input_uri)
 
-        # fill in ID if missing
-        if doc.get('id') is None:
-            doc['id'] = doc_hash
+        with open(input_path) as f:
+            text = to_unicode(f.read())
 
         # pick test/training docs
         if self.options.no_test_set:
             doc['in_test_set'] = False
         else:
+            doc_hash = hashlib.sha1(text.encode('utf-8')).hexdigest()
             doc['in_test_set'] = bool(int(doc_hash[-1], 16) % 2)
 
         # map from (n, ngram) to number of times it appears
         ngram_counts = count_ngrams(
-            doc['text'], self.options.max_ngram_size, self.stop_words)
+            text, self.options.max_ngram_size, self.stop_words)
 
         # yield the number of times the ngram appears in this doc
         # and the categories for this document, so we can train the classifier
         if not doc['in_test_set']:
-            for (n, ngram), count in ngram_counts.iteritems():
+            for (n, ngram), count in ngram_counts.items():
                 yield ('ngram', (n, ngram)), (count, doc['cats'])
 
         # yield the document itself, for safekeeping
-        doc['ngram_counts'] = ngram_counts.items()
+        doc['ngram_counts'] = list(ngram_counts.items())
         yield ('doc', doc['id']), doc
 
     def count_ngram_freq(self, type_and_key, values):
@@ -303,7 +310,7 @@ class MRTextClassifier(MRJob):
 
         for count, cats in values:
             total_df += 1
-            for cat in cats.iteritems():
+            for cat in cats.items():
                 cat_to_df[cat] += 1
                 cat_to_tf[cat] += count
 
@@ -312,7 +319,8 @@ class MRTextClassifier(MRJob):
             return
 
         yield (('global', None),
-               ((n, ngram), (cat_to_df.items(), cat_to_tf.items())))
+               ((n, ngram),
+                (list(cat_to_df.items()), list(cat_to_tf.items()))))
 
     def score_ngrams(self, type_and_key, values):
         """Reducer: Look at all ngrams together, and assign scores by
@@ -372,7 +380,7 @@ class MRTextClassifier(MRJob):
         # rigorous estimate, but it's good enough
         m = len(ngram_to_info)
 
-        for (n, ngram), info in ngram_to_info.iteritems():
+        for (n, ngram), info in ngram_to_info.items():
             # do this even for the special ANY ngram; it's useful
             # as a normalization factor.
             cat_to_df, cat_to_tf = info
@@ -383,7 +391,7 @@ class MRTextClassifier(MRJob):
             # calculate the probability of any given term being
             # this term for documents of each type
             cat_to_p = {}
-            for cat, t in cat_to_t.iteritems():
+            for cat, t in cat_to_t.items():
                 tf = cat_to_tf.get(cat) or 0
                 # use Laplace's rule of succession to estimate p. See:
                 # http://en.wikipedia.org/wiki/Rule_of_succession#Generalization_to_any_number_of_possibilities
@@ -523,18 +531,18 @@ class MRTextClassifier(MRJob):
 
         for (n, ngram), cat_to_score in ngrams_and_scores:
             tf = ngram_counts[(n, ngram)]
-            for cat, score in cat_to_score.iteritems():
+            for cat, score in cat_to_score.items():
                 cat_to_n_to_total_score[cat][n] += score * tf
 
         # average scores for each ngram size
         cat_to_score = {}
-        for cat, n_to_total_score in cat_to_n_to_total_score.iteritems():
+        for cat, n_to_total_score in cat_to_n_to_total_score.items():
             total_score_for_cat = 0
-            for n, total_score in n_to_total_score.iteritems():
+            for n, total_score in n_to_total_score.items():
                 total_t = ngram_counts[(n, None)]
                 total_score_for_cat += (
                     total_score /
-                    max(total_t, self.options.short_doc_threshold, 1))
+                    max(total_t, self.options.short_doc_threshold or 0, 1))
             cat_to_score[cat] = total_score_for_cat
 
         # add scores to the document, and get rid of ngram_counts
@@ -543,5 +551,6 @@ class MRTextClassifier(MRJob):
 
         yield ('doc', doc_id), doc
 
+
 if __name__ == '__main__':
     MRTextClassifier.run()
