@@ -1,5 +1,6 @@
 # Copyright 2009-2010 Yelp
-# Copyright 2015-2016 Yelp
+# Copyright 2015-2018 Yelp
+# Copyright 2019 Yelp
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -20,22 +21,25 @@ Usage::
 
 Options::
 
-  -h, --help            show this help message and exit
-  -c CONF_PATHS, --conf-path=CONF_PATHS
+  -c CONF_PATHS, --conf-path CONF_PATHS
                         Path to alternate mrjob.conf file to read from
   --no-conf             Don't load mrjob.conf even if it's available
-  --emr-endpoint=EMR_ENDPOINT
+  --ec2-endpoint EC2_ENDPOINT
+                        Force mrjob to connect to EC2 on this endpoint (e.g.
+                        ec2.us-west-1.amazonaws.com). Default is to infer this
+                        from region.
+  --emr-endpoint EMR_ENDPOINT
                         Force mrjob to connect to EMR on this endpoint (e.g.
                         us-west-1.elasticmapreduce.amazonaws.com). Default is
                         to infer this from region.
-  --max-days-ago=MAX_DAYS_AGO
+  -h, --help            show this help message and exit
+  --max-days-ago MAX_DAYS_AGO
                         Max number of days ago to look at jobs. By default, we
                         go back as far as EMR supports (currently about 2
                         months)
   -q, --quiet           Don't print anything to stderr
-  --region=REGION       GCE/AWS region to run Dataproc/EMR jobs in.
-  --aws-region=REGION   Deprecated alias for --region
-  --s3-endpoint=S3_ENDPOINT
+  --region REGION       GCE/AWS region to run Dataproc/EMR jobs in.
+  --s3-endpoint S3_ENDPOINT
                         Force mrjob to connect to S3 on this endpoint (e.g. s3
                         -us-west-1.amazonaws.com). You usually shouldn't set
                         this; by default mrjob will choose the correct
@@ -50,22 +54,20 @@ from __future__ import print_function
 import math
 import logging
 import re
+from argparse import ArgumentParser
 from datetime import datetime
 from datetime import timedelta
 from time import sleep
-from optparse import OptionParser
 
+from mrjob.aws import _boto3_now
+from mrjob.aws import _boto3_paginate
 from mrjob.emr import EMRJobRunner
-from mrjob.emr import _list_all_steps
-from mrjob.emr import _yield_all_clusters
-from mrjob.emr import _yield_all_bootstrap_actions
 from mrjob.job import MRJob
-from mrjob.options import _add_basic_options
-from mrjob.options import _add_runner_options
-from mrjob.options import _alphabetize_options
-from mrjob.options import _pick_runner_opts
-from mrjob.parse import iso8601_to_datetime
-from mrjob.patched_boto import _patched_describe_cluster
+from mrjob.options import _add_basic_args
+from mrjob.options import _add_runner_args
+from mrjob.options import _alphabetize_actions
+from mrjob.options import _filter_by_role
+from mrjob.pool import _pool_hash_and_name
 from mrjob.util import strip_microseconds
 
 # match an mrjob job key (used to uniquely identify the job)
@@ -82,16 +84,14 @@ log = logging.getLogger(__name__)
 
 
 def main(args=None):
-    # parser command-line args
-    option_parser = _make_option_parser()
-    options, args = option_parser.parse_args(args)
+    # parse command-line args
+    arg_parser = _make_arg_parser()
 
-    if args:
-        option_parser.error('takes no arguments')
+    options = arg_parser.parse_args(args)
 
     MRJob.set_up_logging(quiet=options.quiet, verbose=options.verbose)
 
-    now = datetime.utcnow()
+    now = _boto3_now()
 
     log.info('getting cluster history...')
     clusters = list(_yield_clusters(
@@ -103,24 +103,25 @@ def main(args=None):
     _print_report(stats, now=now)
 
 
-def _make_option_parser():
-    usage = '%prog [options]'
+def _make_arg_parser():
+    usage = '%(prog)s audit-emr-usage [options]'
     description = 'Print a giant report on EMR usage.'
 
-    option_parser = OptionParser(usage=usage, description=description)
+    arg_parser = ArgumentParser(usage=usage, description=description)
 
-    option_parser.add_option(
-        '--max-days-ago', dest='max_days_ago', type='float', default=None,
+    arg_parser.add_argument(
+        '--max-days-ago', dest='max_days_ago', type=float, default=None,
         help=('Max number of days ago to look at jobs. By default, we go back'
               ' as far as EMR supports (currently about 2 months)'))
 
-    _add_basic_options(option_parser)
-    _add_runner_options(
-        option_parser,
-        _pick_runner_opts('emr', 'connect'))
+    _add_basic_args(arg_parser)
+    _add_runner_args(
+        arg_parser,
+        _filter_by_role(EMRJobRunner.OPT_NAMES, 'connect'))
 
-    _alphabetize_options(option_parser)
-    return option_parser
+    _alphabetize_actions(arg_parser)
+
+    return arg_parser
 
 
 def _runner_kwargs(options):
@@ -132,10 +133,9 @@ def _runner_kwargs(options):
 
 
 def _clusters_to_stats(clusters, now=None):
-    """Aggregate statistics for several clusters into a dictionary.
+    r"""Aggregate statistics for several clusters into a dictionary.
 
-    :param clusters: a sequence of dicts with the keys ``bootstrap_actions``,
-                     ``cluster``, ``steps``.
+    :param clusters: a sequence of dicts with the keys ``cluster``, ``steps``.
     :param now: the current UTC time, as a :py:class:`datetime.datetime`.
                 Defaults to the current time.
 
@@ -256,7 +256,7 @@ def _cluster_to_full_summary(cluster, now=None):
     """Convert a cluster to a full summary for use in creating a report,
     including billing/usage information.
 
-    :param cluster: a :py:class:`boto.emr.EmrObject`
+    :param cluster: a :py:mod:`boto3` cluster data structure
     :param now: the current UTC time, as a :py:class:`datetime.datetime`.
                 Defaults to the current time.
 
@@ -276,11 +276,7 @@ def _cluster_to_full_summary(cluster, now=None):
         cluster, basic_summary=cs, now=now)
 
     # add up billing info
-    if cs['end']:
-        # avoid rounding errors if the job is done
-        cs['nih_billed'] = cs['nih']
-    else:
-        cs['nih_billed'] = float(sum(u['nih_billed'] for u in cs['usage']))
+    cs['nih_billed'] = float(sum(u['nih_billed'] for u in cs['usage']))
 
     for nih_type in ('nih_used', 'nih_bbnu'):
         cs[nih_type] = float(sum(u[nih_type] for u in cs['usage']))
@@ -289,10 +285,9 @@ def _cluster_to_full_summary(cluster, now=None):
 
 
 def _cluster_to_basic_summary(cluster, now=None):
-    """Extract fields such as creation time, owner, etc. from the cluster,
-    so we can safely reference them without using :py:func:`getattr`.
+    """Extract fields such as creation time, owner, etc. from the cluster.
 
-    :param cluster: a :py:class:`boto.emr.EmrObject`
+    :param cluster: a :py:mod:`boto3` cluster data structure
     :param now: the current UTC time, as a :py:class:`datetime.datetime`.
                 Defaults to the current time.
 
@@ -307,7 +302,8 @@ def _cluster_to_basic_summary(cluster, now=None):
       :py:class:`~mrjob.job.MRJob` script that started it), or
       ``None`` for non-:py:mod:`mrjob` clusters.
     * *name*: cluster name, or ``None`` (this should never happen)
-    * *nih*: number of normalized instance hours used by the cluster.
+    * *nih*: number of normalized instance hours cluster *would* use if it
+      ran to the end of the next full hour (
     * *num_steps*: Number of steps in the cluster.
     * *owner*: The owner for the cluster (usually the user that started it),
       or ``None`` for non-:py:mod:`mrjob` clusters.
@@ -321,36 +317,30 @@ def _cluster_to_basic_summary(cluster, now=None):
     * *state*: The cluster's state as a string (e.g. ``'RUNNING'``)
     """
     if now is None:
-        now = datetime.utcnow()
+        now = _boto3_now()
 
     bcs = {}  # basic cluster summary to fill in
 
-    bcs['id'] = getattr(cluster, 'id', None)
-    bcs['name'] = getattr(cluster, 'name', None)
+    bcs['id'] = cluster['Id']
+    bcs['name'] = cluster['Name']
 
-    status = getattr(cluster, 'status', None)
-    timeline = getattr(status, 'timeline', None)
+    Status = cluster['Status']
+    Timeline = Status.get('Timeline', {})
 
-    bcs['created'] = _to_datetime(getattr(
-        timeline, 'creationdatetime', None))
-    bcs['ready'] = _to_datetime(getattr(timeline, 'readydatetime', None))
-    bcs['end'] = _to_datetime(getattr(timeline, 'enddatetime', None))
+    bcs['created'] = Timeline.get('CreationDateTime')
+    bcs['ready'] = Timeline.get('ReadyDateTime')
+    bcs['end'] = Timeline.get('EndDateTime')
 
     if bcs['created']:
         bcs['ran'] = (bcs['end'] or now) - bcs['created']
     else:
         bcs['ran'] = timedelta(0)
 
-    bcs['state'] = getattr(status, 'state', None)
+    bcs['state'] = Status.get('State')
 
-    bcs['num_steps'] = len(getattr(cluster, 'steps', ()))
+    bcs['num_steps'] = len(cluster['Steps'])
 
-    bcs['pool'] = None
-    bootstrap_actions = getattr(cluster, 'bootstrapactions', None)
-    if bootstrap_actions:
-        args = [arg.value for arg in bootstrap_actions[-1].args]
-        if len(args) == 2 and args[0].startswith('pool-'):
-            bcs['pool'] = args[1]
+    _, bcs['pool'] = _pool_hash_and_name(cluster)
 
     m = _JOB_KEY_RE.match(bcs['name'] or '')
     if m:
@@ -358,15 +348,15 @@ def _cluster_to_basic_summary(cluster, now=None):
     else:
         bcs['label'], bcs['owner'] = None, None
 
-    bcs['nih'] = float(getattr(cluster, 'normalizedinstancehours', '0'))
+    bcs['nih'] = float(cluster.get('NormalizedInstanceHours', 0))
 
     return bcs
 
 
 def _cluster_to_usage_data(cluster, basic_summary=None, now=None):
-    """Break billing/usage information for a cluster down by job.
+    r"""Break billing/usage information for a cluster down by job.
 
-    :param cluster: a :py:class:`boto.emr.EmrObject`
+    :param cluster: a :py:mod:`boto3` cluster data structure
     :param basic_summary: a basic summary of the cluster, returned by
                           :py:func:`_cluster_to_basic_summary`. If this
                           is ``None``, we'll call
@@ -405,25 +395,19 @@ def _cluster_to_usage_data(cluster, basic_summary=None, now=None):
     bcs = basic_summary or _cluster_to_basic_summary(cluster)
 
     if now is None:
-        now = datetime.utcnow()
+        now = _boto3_now()
 
     if not bcs['created']:
         return []
 
-    # Figure out billing rate per second for the job, given that
-    # normalizedinstancehours is how much we're charged up until
-    # the next full hour.
-    full_hours = math.ceil(_to_secs(bcs['ran']) / 60.0 / 60.0)
+    # EMR no longer bills by the full hour, but NormalizedInstanceHours
+    # still works that way
+    full_hours = math.ceil(timedelta.total_seconds(bcs['ran']) / 60.0 / 60.0)
     nih_per_sec = bcs['nih'] / (full_hours * 3600.0)
 
-    # Don't actually count a step as billed for the full hour until
-    # the cluster finishes. This means that our total "nih_billed"
-    # will be less than normalizedinstancehours in the cluster, but it
-    # also keeps stats stable for steps that have already finished.
-    if bcs['end']:
-        cluster_end_billing = bcs['created'] + timedelta(hours=full_hours)
-    else:
-        cluster_end_billing = now
+    # EMR bills by the full second, and at least one minute per cluster
+    cluster_end_billing = bcs['created'] + max(
+        _round_up_to_next_second(bcs['ran']), timedelta(minutes=1))
 
     intervals = []
 
@@ -437,17 +421,17 @@ def _cluster_to_usage_data(cluster, basic_summary=None, now=None):
         'step_num': None,
     })
 
-    for step in getattr(cluster, 'steps', ()):
-        step_status = getattr(step, 'status', None)
-        step_timeline = getattr(step_status, 'timeline', None)
+    for step in cluster['Steps']:
+        Status = step['Status']
+        Timeline = Status.get('Timeline', {})
 
         # we've reached the last step that's actually run
-        if not hasattr(step_timeline, 'startdatetime'):
+        if not Timeline.get('StartDateTime'):
             break
 
-        step_start = _to_datetime(step_timeline.startdatetime)
+        step_start = Timeline['StartDateTime']
 
-        step_end = _to_datetime(getattr(step_timeline, 'enddatetime', None))
+        step_end = Timeline.get('EndDateTime')
         if step_end is None:
             # step started running and was cancelled. credit it for 0 usage
             if bcs['end']:
@@ -456,7 +440,7 @@ def _cluster_to_usage_data(cluster, basic_summary=None, now=None):
             else:
                 step_end = now
 
-        m = _STEP_NAME_RE.match(getattr(step, 'name', ''))
+        m = _STEP_NAME_RE.match(step['Name'])
         if m:
             step_label = m.group(1)
             step_owner = m.group(2)
@@ -483,7 +467,7 @@ def _cluster_to_usage_data(cluster, basic_summary=None, now=None):
 
         interval['nih_used'] = (
             nih_per_sec *
-            _to_secs(interval['end'] - interval['start']))
+            timedelta.total_seconds(interval['end'] - interval['start']))
 
         interval['date_to_nih_used'] = dict(
             (d, nih_per_sec * secs)
@@ -498,8 +482,8 @@ def _cluster_to_usage_data(cluster, basic_summary=None, now=None):
                                            interval['end']).items())
 
         interval['nih_billed'] = (
-            nih_per_sec *
-            _to_secs(interval['end_billing'] - interval['start']))
+            nih_per_sec * timedelta.total_seconds(
+                interval['end_billing'] - interval['start']))
 
         interval['date_to_nih_billed'] = dict(
             (d, nih_per_sec * secs)
@@ -538,21 +522,21 @@ def _subdivide_interval_by_date(start, end):
     *start* and *end* are :py:class:`datetime.datetime` objects.
     """
     if start.date() == end.date():
-        date_to_secs = {start.date(): _to_secs(end - start)}
+        date_to_secs = {start.date(): timedelta.total_seconds(end - start)}
     else:
         date_to_secs = {}
 
-        date_to_secs[start.date()] = _to_secs(
-            datetime(start.year, start.month, start.day) + timedelta(days=1) -
-            start)
+        date_to_secs[start.date()] = timedelta.total_seconds(
+            datetime(start.year, start.month, start.day, tzinfo=start.tzinfo) +
+            timedelta(days=1) - start)
 
-        date_to_secs[end.date()] = _to_secs(
-            end - datetime(end.year, end.month, end.day))
+        date_to_secs[end.date()] = timedelta.total_seconds(
+            end - datetime(end.year, end.month, end.day, tzinfo=end.tzinfo))
 
         # fill in dates in the middle
         cur_date = start.date() + timedelta(days=1)
         while cur_date < end.date():
-            date_to_secs[cur_date] = _to_secs(timedelta(days=1))
+            date_to_secs[cur_date] = timedelta.total_seconds(timedelta(days=1))
             cur_date += timedelta(days=1)
 
     # remove zeros
@@ -573,19 +557,20 @@ def _subdivide_interval_by_hour(start, end):
     end_hour = end.replace(minute=0, second=0, microsecond=0)
 
     if start_hour == end_hour:
-        hour_to_secs = {start_hour: _to_secs(end - start)}
+        hour_to_secs = {start_hour: timedelta.total_seconds(end - start)}
     else:
         hour_to_secs = {}
 
-        hour_to_secs[start_hour] = _to_secs(
+        hour_to_secs[start_hour] = timedelta.total_seconds(
             start_hour + timedelta(hours=1) - start)
 
-        hour_to_secs[end_hour] = _to_secs(end - end_hour)
+        hour_to_secs[end_hour] = timedelta.total_seconds(end - end_hour)
 
         # fill in dates in the middle
         cur_hour = start_hour + timedelta(hours=1)
         while cur_hour < end_hour:
-            hour_to_secs[cur_hour] = _to_secs(timedelta(hours=1))
+            hour_to_secs[cur_hour] = timedelta.total_seconds(
+                timedelta(hours=1))
             cur_hour += timedelta(hours=1)
 
     # remove zeros
@@ -606,27 +591,33 @@ def _yield_clusters(max_days_ago=None, now=None, **runner_kwargs):
                           :py:class:`~mrjob.emr.EMRJobRunner`
     """
     if now is None:
-        now = datetime.utcnow()
+        now = _boto3_now()
 
-    emr_conn = EMRJobRunner(**runner_kwargs).make_emr_conn()
+    emr_client = EMRJobRunner(**runner_kwargs).make_emr_client()
 
     # if --max-days-ago is set, only look at recent jobs
     created_after = None
     if max_days_ago is not None:
         created_after = now - timedelta(days=max_days_ago)
 
-    # use _DELAY to sleep 1 second before each API call (see #1091). Could
+    # use _DELAY to sleep 1 second after each API call (see #1091). Could
     # implement some sort of connection wrapper for this if it becomes more
     # generally useful.
-    for cluster_summary in _yield_all_clusters(
-            emr_conn, created_after=created_after, _delay=_DELAY):
-        cluster_id = cluster_summary.id
+    list_clusters_kwargs = dict(_delay=_DELAY)
+    if created_after is not None:
+        list_clusters_kwargs['CreatedAfter'] = created_after
 
+    for cluster_summary in _boto3_paginate(
+            'Clusters', emr_client, 'list_clusters', **list_clusters_kwargs):
+
+        cluster_id = cluster_summary['Id']
+
+        cluster = emr_client.describe_cluster(ClusterId=cluster_id)['Cluster']
         sleep(_DELAY)
-        cluster = _patched_describe_cluster(emr_conn, cluster_id)
-        cluster.steps = _list_all_steps(emr_conn, cluster_id, _delay=_DELAY)
-        cluster.bootstrapactions = list(
-            _yield_all_bootstrap_actions(emr_conn, cluster_id, _delay=_DELAY))
+
+        cluster['Steps'] = list(reversed(list(_boto3_paginate(
+            'Steps', emr_client, 'list_steps',
+            ClusterId=cluster_id, _delay=_DELAY))))
 
         yield cluster
 
@@ -639,7 +630,7 @@ def _print_report(stats, now=None):
                 Defaults to the current time.
     """
     if now is None:
-        now = datetime.utcnow()
+        now = _boto3_now()
 
     s = stats
 
@@ -808,26 +799,6 @@ def _print_report(stats, now=None):
             (cs['owner'] or ''), (cs['label'] or ('not started by mrjob'))))
 
 
-def _to_secs(delta):
-    """Convert a :py:class:`datetime.timedelta` to a number of seconds.
-
-    (This is basically a backport of
-    :py:meth:`datetime.timedelta.total_seconds`.)
-    """
-    return (delta.days * 86400.0 +
-            delta.seconds +
-            delta.microseconds / 1000000.0)
-
-
-def _to_datetime(iso8601_time):
-    """Convert a ISO8601-formatted datetime (from :py:mod:`boto`) to
-    a :py:class:`datetime.datetime`."""
-    if iso8601_time is None:
-        return None
-
-    return iso8601_to_datetime(iso8601_time)
-
-
 def _percent(x, total, default=0.0):
     """Return what percentage *x* is of *total*, or *default* if
     *total* is zero."""
@@ -837,5 +808,13 @@ def _percent(x, total, default=0.0):
         return default
 
 
+def _round_up_to_next_second(td):
+    """Round up to the next second because that's how EMR bills."""
+    if td.microseconds:
+        return strip_microseconds(td) + timedelta(seconds=1)
+    else:
+        return td
+
+
 if __name__ == '__main__':
     main()
