@@ -1,4 +1,5 @@
 # Copyright 2018 Yelp
+# Copyright 2019 Yelp
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -20,6 +21,11 @@ This also adds a few mrjob features that are not standard with
 
 .. versionadded:: 0.6.7
 
+.. versionchanged:: 0.6.8
+
+   added ``local``, ``spark`` runners, made ``spark`` the default (was
+   ``hadoop``)
+
 Usage::
 
     mrjob spark-submit [-r <runner>] [options] <python file | app jar>
@@ -27,113 +33,128 @@ Usage::
 
 Options::
 
-  -r {emr,hadoop}, --runner {emr,hadoop}
-                        Where to run the job (default: "hadoop")
-  --class MAIN_CLASS    Your application's main class (for Java / Scala apps).
-  --name NAME           The name of your application.
-  --jars LIBJARS        Comma-separated list of jars to include on the
-                        driverand executor classpaths.
-  --packages PACKAGES   Comma-separated list of maven coordinates of jars to
-                        include on the driver and executor classpaths. Will
-                        search the local maven repo, then maven central and
-                        any additional remote repositories given by
-                        --repositories. The format for the coordinates should
-                        be groupId:artifactId:version.
-  --exclude-packages EXCLUDE_PACKAGES
-                        Comma-separated list of groupId:artifactId, to exclude
-                        while resolving the dependencies provided in
-                        --packages to avoid dependency conflicts.
-  --repositories REPOSITORIES
-                        Comma-separated list of additional remote repositories
-                        to search for the maven coordinates given with
-                        --packages.
-  --py-files PY_FILES   Comma-separated list of .zip, .egg, or .py files to
-                        placeon the PYTHONPATH for Python apps.
-  --files UPLOAD_FILES  Comma-separated list of files to be placed in the
-                        working directory of each executor. File paths of
-                        these files in executors can be accessed via
-                        SparkFiles.get(fileName).
-  --cmdenv CMDENV       Arbitrary environment variable to set inside Spark, in
-                        the format NAME=VALUE.
-  --conf JOBCONF        Arbitrary Spark configuration property, in the format
-                        PROP=VALUE.
-  --properties-file PROPERTIES_FILE
-                        Path to a file from which to load extra properties. If
-                        not specified, this will look for conf/spark-
-                        defaults.conf.
-  --driver-memory DRIVER_MEMORY
-                        Memory for driver (e.g. 1000M, 2G) (Default: 1024M).
-  --driver-java-options DRIVER_JAVA_OPTIONS
-                        Extra Java options to pass to the driver.
-  --driver-library-path DRIVER_LIBRARY_PATH
-                        Extra library path entries to pass to the driver.
-  --driver-class-path DRIVER_CLASS_PATH
-                        Extra class path entries to pass to the driver. Note
-                        that jars added with --jars are automatically included
-                        in the classpath.
-  --executor-memory EXECUTOR_MEMORY
-                        Memory per executor (e.g. 1000M, 2G) (Default: 1G).
-  --proxy-user PROXY_USER
-                        User to impersonate when submitting the application.
-                        This argument does not work with --principal /
-                        --keytab.
-  -c CONF_PATHS, --conf-path CONF_PATHS
-                        Path to alternate mrjob.conf file to read from
-  --no-conf             Don't load mrjob.conf even if it's available
-  -q, --quiet           Don't print anything to stderr
-  -v, --verbose         print more messages to stderr
-  -h, --help            show this message and exit
-  --master SPARK_MASTER
-                        spark://host:port, mesos://host:port,
-                        yarn,k8s://https://host:port, or local (Default:
-                        yarn).
-  --deploy-mode SPARK_DEPLOY_MODE
-                        Whether to launch the driver program locally
-                        ("client") or on one of the worker machines inside the
-                        cluster ("cluster") (Default: client).
-  --driver-cores DRIVER_CORES
-                        Number of cores used by the driver (Default: 1).
-  --supervise           If given, restarts the driver on failure.
-  --total-executor-cores TOTAL_EXECUTOR_CORES
-                        Total cores for all executors.
-  --executor-cores EXECUTOR_CORES
-                        Number of cores per executor. (Default: 1 in YARN
-                        mode, or all available cores on the worker in
-                        standalone mode)
-  --queue QUEUE_NAME    The YARN queue to submit to (Default: "default").
-  --num-executors NUM_EXECUTORS
-                        Number of executors to launch (Default: 2). If dynamic
-                        allocation is enabled, the initial number of executors
-                        will be at least NUM.
-  --setup SETUP         A command to run before each Spark executor in the
-                        shell ("touch foo"). In cluster mode, runs before the
-                        Spark driver as well. You may interpolate files
-                        available via URL or on your local filesystem using
-                        Hadoop Distributed Cache syntax (". setup.sh#"). To
-                        interpolate archives, use #/: "cd foo.tar.gz#/; make.
-  --archives UPLOAD_ARCHIVES
-                        Comma-separated list of archives to be extracted into
-                        the working directory of each executor.
-  --dirs UPLOAD_DIRS    Comma-separated list of directors to be archived and
-                        then extracted into the working directory of each
-                        executor.
-  --principal PRINCIPAL
-                        Principal to be used to login to KDC, while running
-                        onsecure HDFS.
-  --keytab KEYTAB       The full path to the file that contains the keytab for
-                        the principal specified above. This keytab will be
-                        copied to the node running the Application Master via
-                        the Secure Distributed Cache, for renewing the login
-                        tickets and the delegation tokens periodically.
+ All runners:
+   -r {emr,hadoop,local,spark}, --runner {emr,hadoop,local,spark}
+                         Where to run the job (default: "spark")
+   --class MAIN_CLASS    Your application's main class (for Java / Scala apps).
+   --name NAME           The name of your application.
+   --jars LIBJARS        Comma-separated list of jars to include on the
+                         driverand executor classpaths.
+   --packages PACKAGES   Comma-separated list of maven coordinates of jars to
+                         include on the driver and executor classpaths. Will
+                         search the local maven repo, then maven central and
+                         any additional remote repositories given by
+                         --repositories. The format for the coordinates should
+                         be groupId:artifactId:version.
+   --exclude-packages EXCLUDE_PACKAGES
+                         Comma-separated list of groupId:artifactId, to exclude
+                         while resolving the dependencies provided in
+                         --packages to avoid dependency conflicts.
+   --repositories REPOSITORIES
+                         Comma-separated list of additional remote repositories
+                         to search for the maven coordinates given with
+                         --packages.
+   --py-files PY_FILES   Comma-separated list of .zip, .egg, or .py files to
+                         placeon the PYTHONPATH for Python apps.
+   --files UPLOAD_FILES  Comma-separated list of files to be placed in the
+                         working directory of each executor. Ignored on
+                         local[*] master.
+   --cmdenv CMDENV       Arbitrary environment variable to set inside Spark, in
+                         the format NAME=VALUE.
+   --conf JOBCONF        Arbitrary Spark configuration property, in the format
+                         PROP=VALUE.
+   --setup SETUP         A command to run before each Spark executor in the
+                         shell ("touch foo"). In cluster mode, runs before the
+                         Spark driver as well. You may interpolate files
+                         available via URL or on your local filesystem using
+                         Hadoop Distributed Cache syntax (". setup.sh#"). To
+                         interpolate archives (YARN only), use #/: "cd
+                         foo.tar.gz#/; make.
+   --properties-file PROPERTIES_FILE
+                         Path to a file from which to load extra properties. If
+                         not specified, this will look for conf/spark-
+                         defaults.conf.
+   --driver-memory DRIVER_MEMORY
+                         Memory for driver (e.g. 1000M, 2G) (Default: 1024M).
+   --driver-java-options DRIVER_JAVA_OPTIONS
+                         Extra Java options to pass to the driver.
+   --driver-library-path DRIVER_LIBRARY_PATH
+                         Extra library path entries to pass to the driver.
+   --driver-class-path DRIVER_CLASS_PATH
+                         Extra class path entries to pass to the driver. Note
+                         that jars added with --jars are automatically included
+                         in the classpath.
+   --executor-memory EXECUTOR_MEMORY
+                         Memory per executor (e.g. 1000M, 2G) (Default: 1G).
+   --proxy-user PROXY_USER
+                         User to impersonate when submitting the application.
+                         This argument does not work with --principal /
+                         --keytab.
+   -c CONF_PATHS, --conf-path CONF_PATHS
+                         Path to alternate mrjob.conf file to read from
+   --no-conf             Don't load mrjob.conf even if it's available
+   -q, --quiet           Don't print anything to stderr
+   -v, --verbose         print more messages to stderr
+   -h, --help            show this message and exit
+
+ Spark and Hadoop runners only:
+   --master SPARK_MASTER
+                         spark://host:port, mesos://host:port,
+                         yarn,k8s://https://host:port, or local. Defaults to
+                         local[*] on spark runner, yarn on hadoop runner.
+   --deploy-mode SPARK_DEPLOY_MODE
+                         Whether to launch the driver program locally
+                         ("client") or on one of the worker machines inside the
+                         cluster ("cluster") (Default: client).
+
+ Cluster deploy mode only:
+   --driver-cores DRIVER_CORES
+                         Number of cores used by the driver (Default: 1).
+
+ Spark standalone or Mesos with cluster deploy mode only:
+   --supervise           If given, restarts the driver on failure.
+
+ Spark standalone and Mesos only:
+   --total-executor-cores TOTAL_EXECUTOR_CORES
+                         Total cores for all executors.
+
+ Spark standalone and YARN only:
+   --executor-cores EXECUTOR_CORES
+                         Number of cores per executor. (Default: 1 in YARN
+                         mode, or all available cores on the worker in
+                         standalone mode)
+
+ YARN-only:
+   --archives UPLOAD_ARCHIVES
+                         Comma-separated list of archives to be extracted into
+                         the working directory of each executor.
+   --dirs UPLOAD_DIRS    Comma-separated list of directors to be archived and
+                         then extracted into the working directory of each
+                         executor.
+   --queue QUEUE_NAME    The YARN queue to submit to (Default: "default").
+   --num-executors NUM_EXECUTORS
+                         Number of executors to launch (Default: 2). If dynamic
+                         allocation is enabled, the initial number of executors
+                         will be at least NUM.
+   --principal PRINCIPAL
+                         Principal to be used to login to KDC, while running
+                         onsecure HDFS.
+   --keytab KEYTAB       The full path to the file that contains the keytab for
+                         the principal specified above. This keytab will be
+                         copied to the node running the Application Master via
+                         the Secure Distributed Cache, for renewing the login
+                         tickets and the delegation tokens periodically.
 
 This also supports the same runner-specific switches as
 :py:class:`~mrjob.job.MRJob`\\s (e.g. ``--hadoop-bin``, ``--region``).
+
 """
 from __future__ import print_function
 
 import os
 import sys
 from argparse import ArgumentParser
+from argparse import REMAINDER
 from argparse import SUPPRESS
 from logging import getLogger
 
@@ -168,17 +189,17 @@ _DEPRECATED_OPT_HELP = (
 # display in default help message
 
 # the only runners that support spark scripts/jars
-_SPARK_RUNNERS = ('emr', 'hadoop')
+_SPARK_RUNNERS = ('emr', 'hadoop', 'local', 'spark')
 
 # the default spark runner to use
-_DEFAULT_RUNNER = 'hadoop'
+_DEFAULT_RUNNER = 'spark'  # just find spark-submit and use it
 
 
 # our mostly similar version of spark-submit's args, arranged in to groups
 # for the --help message. Differences:
 #
-# spark_master (--master) is in its own "Hadoop runner only" group
-# added upload_dirs (--dirs) which is similar to --archive
+# spark_master (--master) is in its own "Spark and Hadoop runners only" group
+# added upload_dirs (--dirs) which is similar to --archives
 #
 # --runner and other basic options are patched into the first ("None")
 # argument group in _make_basic_help_parser(), below
@@ -194,6 +215,7 @@ _SPARK_SUBMIT_ARG_GROUPS = [
         'upload_files',
         'cmdenv',
         'jobconf',
+        'setup',
         'properties_file',
         'driver_memory',
         'driver_java_options',
@@ -202,7 +224,7 @@ _SPARK_SUBMIT_ARG_GROUPS = [
         'executor_memory',
         'proxy_user',
     ]),
-    ('Hadoop runner only', [
+    ('Spark and Hadoop runners only', [
         'spark_master',
         'spark_deploy_mode',
     ]),
@@ -220,7 +242,6 @@ _SPARK_SUBMIT_ARG_GROUPS = [
         'executor_cores',
     ]),
     ('YARN-only', [
-        'setup',
         'upload_archives',
         'upload_dirs',
         'queue_name',
@@ -284,12 +305,14 @@ _SPARK_SUBMIT_ARG_HELP = dict(
            ' driver as well. You may interpolate files'
            ' available via URL or on your local filesystem using'
            ' Hadoop Distributed Cache syntax (". setup.sh#"). To'
-           ' interpolate archives, use #/: "cd foo.tar.gz#/; make.'),
+           ' interpolate archives (YARN only), use'
+           ' #/: "cd foo.tar.gz#/; make.'),
     spark_deploy_mode=('Whether to launch the driver program locally'
                        ' ("client") or on one of the worker machines inside'
                        ' the cluster ("cluster") (Default: client).'),
     spark_master=('spark://host:port, mesos://host:port, yarn,'
-                  'k8s://https://host:port, or local (Default: yarn).'),
+                  'k8s://https://host:port, or local. Defaults'
+                  ' to local[*] on spark runner, yarn on hadoop runner.'),
     supervise='If given, restarts the driver on failure.',
     total_executor_cores='Total cores for all executors.',
     upload_archives=('Comma-separated list of archives to be extracted into'
@@ -297,9 +320,7 @@ _SPARK_SUBMIT_ARG_HELP = dict(
     upload_dirs=('Comma-separated list of directors to be archived and then'
                  ' extracted into the working directory of each executor.'),
     upload_files=('Comma-separated list of files to be placed in the working'
-                  ' directory of each executor. File paths of these files'
-                  ' in executors can be accessed via'
-                  ' SparkFiles.get(fileName).'),
+                  ' directory of each executor. Ignored on local[*] master.'),
 )
 
 _SPARK_SUBMIT_OPT_NAMES = {
@@ -465,7 +486,7 @@ def _make_arg_parser():
 
     # add positional arguments
     parser.add_argument(dest='script_or_jar', nargs='?')
-    parser.add_argument(dest='args', nargs='*')
+    parser.add_argument(dest='args', nargs=REMAINDER)
 
     _add_basic_args(parser)
     _add_runner_alias_arg(parser)
